/*      file     : ParSim/TParSimGraphics.C
        project  : PAMINA III
        author   : Marcus Rickert
        RCS      : $Id: TParSimGraphics.C,v 1.4 2005/03/25 11:07:34 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <TParallel.h>
#include <TConfiguration.h>
#include <TTimeStep.h>
#include <TNode.h>
#include <TEdge.h>
#include <TLocationInfo.h>
#include <TAdjacenceInfo.h>
#include <TObjectManager.h>
#include <TTrafficDataManager.h>
#include <TMultilaneDump.h>

#include <ParSimConfig.h>

#ifdef XWINDOWS

#  include <TTopologyWindow.h>
#  include <TMapViewer.h>
#  include <TGraphEdgeInfo.h>

#  include "TParSimGraphics.h"

#endif

// ***DEFINES************************************************************************* 

#define PARSIM_DEFAULT_TOPOLOGYWINDOW 1L
#define PARSIM_DEFAULT_MAPVIEWER      1L

// ***CONSTANTS***********************************************************************
// ***TYPES*************************************************************************** 
// ***GLOBAL VARIABLES****************************************************************  
// ***LOCAL VARIABLES***************************************************************** 
// ***FUNCTIONS*********************************************************************** 
// ***METHODS************************************************************************* 

#ifdef NO_INLINE
#endif

// =================================================================================== 

// =================================================================================== 
#define ParentClass TBaseObject
// =================================================================================== 

TParSimGraphics::TParSimGraphics(void)

{
    TopologyWindowActive = Configuration->Exists(PARSIM_GROUP, PARSIM_NAME_TOPOLOGYWINDOW)? 
        Configuration->GetLong() : PARSIM_DEFAULT_TOPOLOGYWINDOW;
    MapViewerActive = Configuration->Exists(PARSIM_GROUP, PARSIM_NAME_MAPVIEWER)? 
        Configuration->GetLong() : PARSIM_DEFAULT_MAPVIEWER;

    TopologyWindow = 0;
    MapViewer = 0;
    TopologyChanged = TRUE;
    Nodes = new TNodeList;
    Edges = new TEdgeList;
}

// =================================================================================== 

TParSimGraphics::~TParSimGraphics(void)

{
    DeleteObject(Nodes);
    DeleteObject(Edges);
    DeleteObject(MapViewer);
    DeleteObject(TopologyWindow);
}

// =================================================================================== 

bool
TParSimGraphics::Activate(void)

{
    Print(DM_LIFESPAN, "Activate");
    return TRUE;
}

// =================================================================================== 

bool
TParSimGraphics::ExecuteSubTimeStep(TTimeStep &TimeStep)

{
    if (TimeStep.GetSubStep() == 0)
    {
        if (TopologyWindow)
            if (!TopologyWindow->ExecuteSubTimeStep(TimeStep))
            {
                Error("ExecuteSubTimeStep: error while handling topology window");
                return FALSE;
            }
        if (MapViewer)
        {
            if (TopologyChanged)
                MapViewer->MarkViews();
            if (!MapViewer->ExecuteSubTimeStep(TimeStep))
            {
                Error("ExecuteSubTimeStep: error while handling map viewer");
                return FALSE;
            }
        }
        TopologyChanged = FALSE;
    }
    return TRUE;
}

// =================================================================================== 

TString
TParSimGraphics::GetClassName(void) const

{
    return "TParSimGraphics";
}

// =================================================================================== 

bool
TParSimGraphics::HandleAction(TParallelObject *Object, TObjectAction Action)

{
    if (Object->GetClassID() == TNode::ClassInfo->GetID())
    {
        switch (Action)
        {
        case ActionMigrated:

            if (MapViewer)
            {           
                if (!MapViewer->UpdateNode((TNode *) Object))
                {
                    Error("HandleAction: cannot update %s", 
                          (char *) Object->GetObjectName());
                    return FALSE;
                }
            }
            TopologyChanged = TRUE;
            return TRUE;

        default:
            ;
        }
    } 
    return TRUE;
}

// =================================================================================== 

bool
TParSimGraphics::LoadMap(void)

{
    TInfoMask NodeLocalMask;
    TInfoMask EdgeLocalMask;
    TInfoMask NodeLoadMask;
    TInfoMask EdgeLoadMask;
    TNodeEnum CNode;
    TInfoID   LocationInfoID;

    NodeLocalMask = TLocationInfo::ClassInfo->GetInfoMask();
    EdgeLocalMask = (TMultilaneDump::ClassInfo->GetInfoMask() |
                     TAdjacenceInfo::ClassInfo->GetInfoMask());
#ifdef XWINDOWS
    EdgeLocalMask |= TGraphEdgeInfo::ClassInfo->GetInfoMask();
#endif
    if (!TrafficDataManager->CreateSimulationObjects(Nodes, Edges, 
                                                     NodeLocalMask, EdgeLocalMask))
    {
        Error("LoadMap: cannot create graphic objects");
        return FALSE;
    }
    NodeLoadMask = TLocationInfo::ClassInfo->GetInfoMask();
    EdgeLoadMask = TAdjacenceInfo::ClassInfo->GetInfoMask();
    if (!(ObjectManager->FillContainers((TContainerObjectList *) Nodes, NodeLoadMask) AND
          ObjectManager->FillContainers((TContainerObjectList *) Edges, EdgeLoadMask)))
    {
        Error("cannot fill graphic objects");
        return FALSE;
    }
    LocationInfoID = TLocationInfo::ClassInfo->GetInfoID();
    Frame.Empty();
    CNode.WalksForwardThrough(Nodes);
    while (CNode())
    {
        Frame.Merge(CNode->Loc(LocationInfoID));
        CNode++;
    }
    Frame *= 1.1;
    return TRUE;
}

// =================================================================================== 

bool
TParSimGraphics::Run(void)

{
    Print(DM_LIFESPAN, "Run");
    if (!LoadMap())
    {
        Error("Run: cannot load map");
        return FALSE;
    }
    if (TopologyWindowActive)
        TopologyWindow = new TTopologyWindow(Parallel->GetPhysicalTopology());
    if (MapViewerActive)
    {
        MapViewer = new TMapViewer(&Frame, Nodes, Edges);
        TNode::ClassInfo->RegisterActionHandler(this);
    }
    return TRUE;
}

// =================================================================================== 
#undef ParentClass
