/*      file     : TrafficData/TIntersection.C
        project  : PAMINA III
        author   : Marcus Rickert
        RCS      : $Id: TIntersection.C,v 1.9 2005/03/25 12:28:51 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <stdio.h>
#include <stdlib.h>

#include <TString.h>
#include <TBaseStream.h>
#include <TTransferInfo.h>
#include <TIncidenceInfo.h>
#include <TTimeStep.h>
#include <TEdge.h>

#include "TSource.h"
#include "TSink.h"
#include "TMultilaneGrid.h"
#include "TRouteplan.h"
#include "TIntersectionControl.h"
#include "TIntersection.h"

// ***METHODS************************************************************************* 

#ifdef NO_INLINE
#  include <TIntersection.i.C>
#endif

// =================================================================================== 

TIntersectionVarData::TIntersectionVarData(void)

{
    RoutesFailed = 0;
}

// =================================================================================== 
#define ParentClass TParallelObject
// =================================================================================== 

TIntersection::TIntersection(TClassInfo *ClassInfo,TParallelID theParallelID) :

ParentClass(ClassInfo, theParallelID)

{
    TurnCounts = 0;
    IncomingGrids = 0;
    OutgoingGrids = 0;
    Sources = 0;
    IntersectionControl = 0;
    Sink = new TSink(theParallelID);
}

// ===================================================================================

TIntersection::~TIntersection(void)

{
    DeleteObjectArray(TurnCounts);
    DeleteObjectArray(IncomingGrids);
    DeleteObjectArray(OutgoingGrids);
    DeleteObjectArray(Sources);
    DeleteObject(Sink);
}

// ===================================================================================

bool
TIntersection::Activate(void)

{
    TIntersectionControl *Control;

    IntersectionControl = 0;
    if ((Control = (TIntersectionControl *) 
         GetInfo(TIntersectionControl::ClassInfo->GetInfoID())))
    {
        if (Control->GetNrOfPhases() > 0)
            IntersectionControl = Control;
    }
    return TRUE;
}

// ===================================================================================

bool
TIntersection::AddVehicle(TVehicle *Vehicle)

{
    TParallelID EdgeID;
    int         GridNr;

    EdgeID = Vehicle->GetRouteplan()->GetCurrentID();
    if ((GridNr = GetOutgoingGridNrByID(EdgeID)) == INVALID)
    {
        static char Format[128] = { 0 };

        if (0 == Format[0]) {
            sprintf(Format, "AddVehicle: cannot find edge %ld for route %%s", 
                    PARALLEL_ID_PRINTF_FORMAT);
        }
        Error(Format, EdgeID, (char *) Vehicle->GetRouteplan()->GetInstanceName());
        return FALSE;
    }
    if (Sources[GridNr] == 0)
    {
        Sources[GridNr] = new TSource(GetParallelID());
        Sources[GridNr]->AssignGrid(OutgoingGrids[GridNr], 0);
    }
    return Sources[GridNr]->AddVehicle(Vehicle);
}

// ===================================================================================

bool
TIntersection::AssignEdges(TIncidenceInfo *theIncidenceInfo)

{
    TMultilaneGrid *Grid;
    TEdge          *Edge;
    int             CEdge;
    TInfoID         InfoID;

    //    Print("assign %s", (char *) theIncidenceInfo->GetObjectName());

    IncidenceInfo = theIncidenceInfo;
    InfoID = TMultilaneGrid::ClassInfo->GetInfoID();
    if (IncidenceInfo->GetNrOfIncomingEdges())
        IncomingGrids = new TMultilaneGrid * [IncidenceInfo->GetNrOfIncomingEdges()];
    for (CEdge = 0; CEdge < IncidenceInfo->GetNrOfIncomingEdges(); CEdge++)
    {
        Edge = IncidenceInfo->GetIncomingEdge(CEdge);
        if ((Grid = (TMultilaneGrid *) Edge->GetInfo(InfoID)) == 0)
        {
            Error("AssignEdges: cannot retrieve grid for incoming edge %s", 
                  (char *) Edge->GetInstanceName());
            return FALSE;
        }
        IncomingGrids[CEdge] = Grid;
    }
    if (IncidenceInfo->GetNrOfOutgoingEdges())
    {
        OutgoingGrids = new TMultilaneGrid * [IncidenceInfo->GetNrOfOutgoingEdges()];
        Sources = new TSource * [IncidenceInfo->GetNrOfOutgoingEdges()];
    }
    for (CEdge = 0; CEdge < IncidenceInfo->GetNrOfOutgoingEdges(); CEdge++)
    {
        Edge = IncidenceInfo->GetOutgoingEdge(CEdge);
        if ((Grid = (TMultilaneGrid *) Edge->GetInfo(InfoID)) == 0)
        {
            Error("AssignEdges: cannot retrieve grid for outgoing edge %s", 
                  (char *) Edge->GetInstanceName());
            return FALSE;
        }
        OutgoingGrids[CEdge] = Grid;
        Sources[CEdge] = 0;
    }
    if (IncidenceInfo->GetNrOfIncomingEdges() == 0 AND
        IncidenceInfo->GetNrOfOutgoingEdges() == 0)
        Warning("AssignEdges: intersection is isolated");
    TurnCounts = new long [IncidenceInfo->GetNrOfIncomingEdges() * 
                          IncidenceInfo->GetNrOfOutgoingEdges()];
    ResetTurnCounts();
    return TRUE;
}

// ===================================================================================

bool
TIntersection::Decode(TBaseStream *aStream, TTransferInfo *TransferInfo)

{
    if (!ParentClass::Decode(aStream, TransferInfo))
        return FALSE;
    switch(TransferInfo->TransferMode)
    {
    case TM_ConstData:
        return TRUE;

    case TM_VarData:
        return aStream->Decode(Var.RoutesFailed);

    default:
        Warning("Decode: invalid TransferMode %d", TransferInfo->TransferMode);
    }
    return TRUE;
}

// ===================================================================================

bool
TIntersection::Encode(TBaseStream *aStream, TTransferInfo *TransferInfo) const

{
    if (!ParentClass::Encode(aStream, TransferInfo))
        return FALSE;
    switch(TransferInfo->TransferMode)
    {
    case TM_ConstData:
        return TRUE;

    case TM_VarData:
        return aStream->Encode(Var.RoutesFailed);

    default:
        Warning("Encode: invalid TransferMode %d", TransferInfo->TransferMode);
    }
    return TRUE;
}

// =================================================================================== 

bool
TIntersection::ExecuteSubTimeStep(const TTimeStep &TimeStep)

{
    int CSource;

    IfDebug(DM_TIME_STEP, Print("ExecuteSubTimeStep: %s", (char *) TimeStep.GetInstanceName()));
    for (CSource = 0; CSource < IncidenceInfo->GetNrOfOutgoingEdges(); CSource++)
        if (Sources[CSource])
            if (!Sources[CSource]->ExecuteSubTimeStep(TimeStep))
            {
                Error("ExecuteSubTimeStep: while executing source");
                return FALSE;
            }
    return TRUE;
}

// ===================================================================================

TString
TIntersection::GetClassName(void) const

{
    return "TIntersection";
}

// =================================================================================== 

TMultilaneGrid *
TIntersection::GetOutgoingGridByID(TParallelID EdgeID) const
    
{
    int CGrid;

    for (CGrid = 0; CGrid < IncidenceInfo->GetNrOfOutgoingEdges(); CGrid++)
        if (OutgoingGrids[CGrid]->GetParallelID() == EdgeID)
            return OutgoingGrids[CGrid];
    return 0;
}

// =================================================================================== 

int
TIntersection::GetOutgoingGridNrByID(TParallelID EdgeID) const
    
{
    int CGrid;

    for (CGrid = 0; CGrid < IncidenceInfo->GetNrOfOutgoingEdges(); CGrid++)
        if (OutgoingGrids[CGrid]->GetParallelID() == EdgeID)
            return CGrid;

    char Format1[128] = { 0 };
    char Format2[128];

    if (0 == Format1[0]) {
        sprintf(Format1, "WARNING: looking for edge %s, but ...", PARALLEL_ID_PRINTF_FORMAT);
        sprintf(Format2, "...found %s", PARALLEL_ID_PRINTF_FORMAT);
    }
    Print(Format1, EdgeID);
    for (CGrid = 0; CGrid < IncidenceInfo->GetNrOfOutgoingEdges(); CGrid++)
        Print(Format2, OutgoingGrids[CGrid]->GetParallelID());
    return INVALID;
}

// ===================================================================================

TSource *
TIntersection::GetSource(TParallelID GridID) const

{
    int GridNr;

    if (Sources == 0)
        return 0;
    if ((GridNr = GetOutgoingGridNrByID(GridID)) == INVALID)
        return 0;
    return Sources[GridNr];
}

// ===================================================================================

long 
TIntersection::GetTurnCount(int FromGridNr, int ToGridNr) const

{
    if (TurnCounts == 0)
        return 0;
    return TurnCounts[FromGridNr * IncidenceInfo->GetNrOfOutgoingEdges() + ToGridNr];
}

// =================================================================================== 

void
TIntersection::ResetTurnCounts(void)

{
    int i, Count;

    Count = IncidenceInfo->GetNrOfIncomingEdges() * IncidenceInfo->GetNrOfOutgoingEdges();
    for (i = 0; i < Count; i++)
        TurnCounts[i] = 0;
}

// =================================================================================== 

bool
TIntersection::Turn(int FromGridNr, int ToGridNr)

{
    int Index;
    if (TurnCounts == 0)
        TurnCounts = new long [IncidenceInfo->GetNrOfIncomingEdges() * 
                              IncidenceInfo->GetNrOfOutgoingEdges()];
    Index = FromGridNr * IncidenceInfo->GetNrOfOutgoingEdges() + ToGridNr;
    TurnCounts[Index]++;
    return TRUE;
}

// =================================================================================== 

void
TIntersection::VehicleFailed(TVehicle *Vehicle)

{
    Sink->VehicleFailed(Vehicle);
}

// =================================================================================== 
#undef ParentClass


