/*      file     : TrafficData/TMultilaneStat.C
        project  : PAMINA III
        author   : Marcus Rickert
        RCS      : $Id: TMultilaneStat.C,v 1.8 2005/03/25 12:28:52 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

/* ***INCLUDES************************************************************************ */

#include <math.h>

#include <TBaseStream.h>
#include <TTimeStep.h>
#include "TMultilaneGrid.h"
#include "TMultilaneStat.h"

// ***GLOBAL VARIABLES****************************************************************  

TClassInfo *TMultilaneStat::ClassInfo = 0;

/* ***METHODS************************************************************************* */

#ifdef NO_INLINE
#  include <TMultilaneStat.i.C>
#endif

/* =================================================================================== */
#define ParentClass TStatObject
/* =================================================================================== */

TMultilaneStat::TMultilaneStat(TParallelID theParallelID) :

ParentClass(ClassInfo, theParallelID)

{
    Reset();
}

/* =================================================================================== */

TMultilaneStat::~TMultilaneStat(void)

{
}

/* =================================================================================== */

/* The method 'Add' must be overwritten to ensure that two statistic objects (here
   the current object and the object passed as a parameter) can
   be merged into one single object (which is the current object) ususally by
   adding or multiplying local variables. But of course any calculation can be
   performed that is both commutative and associative. */

bool
TMultilaneStat::Accumulate(const TStatObject *StatObject)

{
    const TMultilaneStat *Stat = (const TMultilaneStat *) StatObject;

    int CLane;
    int CV;

    NrOfGrids += Stat->NrOfGrids;
    NrOfActiveGrids += Stat->NrOfActiveGrids;
    NrOfActiveSites += Stat->NrOfActiveSites;
    if (Stat->MaxNrOfLanes > MaxNrOfLanes)
        MaxNrOfLanes = Stat->MaxNrOfLanes;
    for (CLane = 0; CLane < MaxNrOfLanes; CLane++)
    {
        NrOfSites[CLane] += Stat->NrOfSites[CLane];
        NrOfSlowVehicles[CLane] += Stat->NrOfSlowVehicles[CLane];
        for (CV = 0; CV <= CA_DEFAULT_V_MAX; CV++)
            V[CLane][CV] += Stat->V[CLane][CV];
    }
    return ParentClass::Accumulate(StatObject);
}


/* =================================================================================== */

bool
TMultilaneStat::Collect(const TMultilaneGrid *Grid)

{
    long CLane, CSite;

    NrOfGrids++;
    NrOfActiveSites += Grid->GetNrOfLocalSites();
    if (Grid->GetNrOfActiveSites() > 0)
        NrOfActiveGrids++;
    if (Grid->GetNrOfLanes() > MaxNrOfLanes)
        MaxNrOfLanes = Grid->GetNrOfLanes();
    for (CLane = 0; CLane < Grid->GetNrOfLanes(); CLane++)
    {
        NrOfSites[CLane] += Grid->GetNrOfLocalSites();
        for (CSite = Grid->GetFirstSite(); 
             CSite < Grid->GetLastSite() + Grid->GetBndWidth(); CSite++)
            if (Grid->VAccess(CLane, CSite) != 0)
            {
                V[CLane][SITE_GET_VELOCITY(Grid->SAccess(CLane, CSite))]++;
                if (SITE_GET_V_MAX(Grid->SAccess(CLane, CSite)) < CA_DEFAULT_V_MAX)
                    NrOfSlowVehicles[CLane]++;
            }
    }
    return TRUE;
}

/* =================================================================================== */

bool
TMultilaneStat::Decode(TBaseStream *aStream, TTransferInfo *TransferInfo)

{
    if (!(ParentClass::Decode(aStream, TransferInfo) AND
          aStream->Decode(MaxNrOfLanes) AND
          aStream->Decode(NrOfActiveSites) AND
          aStream->Decode(NrOfGrids) AND
          aStream->Decode(NrOfActiveGrids)))
        return FALSE;
    if (!aStream->Decode(NrOfSites[0], (MAX_LANES + 1) * LONGINT_PER_LANE))
        return FALSE;
    return TRUE;
}

/* =================================================================================== */

bool
TMultilaneStat::Encode(TBaseStream *aStream, TTransferInfo *TransferInfo) const

{
    if (!(ParentClass::Encode(aStream, TransferInfo) AND
          aStream->Encode(MaxNrOfLanes) AND
          aStream->Encode(NrOfActiveSites) AND
          aStream->Encode(NrOfGrids) AND
          aStream->Encode(NrOfActiveGrids)))
        return FALSE;
    if (!aStream->Encode(NrOfSites[0], (MAX_LANES + 1) * LONGINT_PER_LANE))
        return FALSE;
    return TRUE;
}

/* =================================================================================== */

bool
TMultilaneStat::Evaluate(void)

{
    int  CV;
    int  CLane;
    long Temp;

    for (CV = 0; CV <= CA_DEFAULT_V_MAX; CV++)
    {
        V[ALL_LANES][CV] = 0;
        for (CLane = 0; CLane < MAX_LANES; CLane++)
            V[ALL_LANES][CV] += V[CLane][CV];
    }
    V[ALL_LANES][ALL_V] = 0;
    SumV[ALL_LANES] = 0;
    SumV2[ALL_LANES] = 0;
    NrOfSites[ALL_LANES] = 0;
    NrOfSlowVehicles[ALL_LANES] = 0;
    for (CLane = 0; CLane < MAX_LANES; CLane++)
    {
        SumV[CLane] = 0;
        SumV2[CLane] = 0;
        V[CLane][ALL_V] = 0;
        for (CV = 0; CV <= CA_DEFAULT_V_MAX; CV++)
        {
            V[CLane][ALL_V] += V[CLane][CV];
            SumV[CLane] += (Temp = CV * V[CLane][CV]);
            SumV2[CLane] += CV * Temp;
        }
        V[ALL_LANES][ALL_V] += V[CLane][ALL_V];
        SumV[ALL_LANES] += SumV[CLane];
        SumV2[ALL_LANES] += SumV2[CLane];
        NrOfSites[ALL_LANES] += NrOfSites[CLane];
        NrOfSlowVehicles[ALL_LANES] += NrOfSlowVehicles[CLane];
    }
    return TRUE;
}

/* =================================================================================== */

TString
TMultilaneStat::GetClassName(void) const

{
    return "TMultilaneStat";
}

/* =================================================================================== */

double
TMultilaneStat::GetVDelta(int CLane) const

{
    if (GetNrOfVehicles(CLane))
        return sqrt((double) SumV2[CLane] * (double) GetNrOfVehicles(CLane) -
                    (double) SumV[CLane] * (double) SumV[CLane]) / (double) GetNrOfVehicles(CLane);
    else
        return INVALID_MEAN;
}

/* =================================================================================== */

void
TMultilaneStat::PrintLineStat(TTimeStep RefTimeStep, FILE *OutputFile)

{
    int CLane;

    fprintf(OutputFile, "%10ld %10ld %10.8f %10.8f %10ld %10.8f %10.8f %10.8f",
            (long) RefTimeStep, GetNrOfSites(), GetDensity(), GetSlowRatio(),
            GetSumV(), GetVMean(), GetVDelta(), GetFlow());
    for (CLane = 0; CLane < MaxNrOfLanes; CLane++)
        fprintf(OutputFile, " %10ld %10.8f %10.8f %10ld %10.8f %10.8f %10.8f",
                GetNrOfSites(CLane), GetDensity(CLane), GetSlowRatio(),
                GetSumV(CLane), GetVMean(CLane), GetVDelta(CLane), GetFlow(CLane));
    fprintf(OutputFile, "\n");
}

/* =================================================================================== */

void
TMultilaneStat::PrintLineStatDescription(FILE *OutputFile)

{
    int CLane;
    int Col;

    fprintf(OutputFile, "# col: description\n");
    Col = 0;
    fprintf(OutputFile, "# %2d : timestep\n", ++Col);
    fprintf(OutputFile, "# %2d : sites (all)\n", ++Col);
    fprintf(OutputFile, "# %2d : density (all)\n", ++Col);
    fprintf(OutputFile, "# %2d : slow ratio (all)\n", ++Col);
    fprintf(OutputFile, "# %2d : sum v (all)\n", ++Col);
    fprintf(OutputFile, "# %2d : mean v (all)\n", ++Col);
    fprintf(OutputFile, "# %2d : delta v (all)\n", ++Col);
    fprintf(OutputFile, "# %2d : flow (all)\n", ++Col);
    for (CLane = 0; CLane < MaxNrOfLanes; CLane++)
    {
        fprintf(OutputFile, "# %2d : sites (lane %d)\n", ++Col, CLane);
        fprintf(OutputFile, "# %2d : density (lane %d)\n", ++Col, CLane);
        fprintf(OutputFile, "# %2d : slow ratio (lane %d)\n", ++Col, CLane);
        fprintf(OutputFile, "# %2d : sum v (lane %d)\n", ++Col, CLane);
        fprintf(OutputFile, "# %2d : mean v (lane %d)\n", ++Col, CLane);
        fprintf(OutputFile, "# %2d : delta v (lane %d)\n", ++Col, CLane);
        fprintf(OutputFile, "# %2d : flow (lane %d)\n", ++Col, CLane);
    }
    fprintf(OutputFile, "\n\n");    
}

/* =================================================================================== */

void
TMultilaneStat::PrintScreenStat(FILE *OutputFile)

{
    int CLane;

    fprintf(OutputFile, "Statistics of time step %ld (+%dh%02d')\n", (long) GetTimeStep(),
            GetTimeStep().Hours(), GetTimeStep().Minutes());    
    fprintf(OutputFile, "Lane     Sites  Vehicles Dens      SumV  VMean   Flow\n");
    for (CLane = 0; CLane < MaxNrOfLanes; CLane++)
        fprintf(OutputFile, "%4d%10ld%10ld %4.2f%10ld %6.4f %6.4f\n", CLane+1,
                GetNrOfSites(CLane), GetNrOfVehicles(CLane),
                GetDensity(CLane),
                GetSumV(CLane), GetVMean(CLane), GetFlow(CLane));
    fprintf(OutputFile, " ALL%10ld%10ld %4.2f%10ld %6.4f %6.4f\n", GetNrOfSites(), 
            GetNrOfVehicles(), GetDensity(), GetSumV(), GetVMean(), GetFlow());
}

/* =================================================================================== */

void
TMultilaneStat::Reset(void)

{
    long CLane;
    int  CV;

    ParentClass::Reset();
    MaxNrOfLanes = 0;
    NrOfGrids = 0;
    NrOfActiveGrids = 0;
    NrOfActiveSites = 0;
    for (CLane = 0; CLane < MAX_LANES; CLane++)
    {
        NrOfSites[CLane] = 0;
        NrOfSlowVehicles[CLane] = 0;
        for (CV = 0; CV <= CA_DEFAULT_V_MAX; CV++)
            V[CLane][CV] = 0;
    }   
}

/* =================================================================================== */
#undef ParentClass
