/*      file     : partools/TParallelObject.C
        project  : Parallel Toolbox 2.x
        author   : Marcus Rickert
        RCS      : $Id: TParallelObject.C,v 1.16 2005/03/25 11:08:15 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <malloc.h>
#include <stdio.h>

#include <TBaseMessage.h>

#include "TClassInfo.h"
#include "TTransferInfo.h"
#include "TContainerObject.h"
#include "TParallelObject.h"

// ***LOCAL VARIABLES***************************************************************** 

#ifndef NO_FAST_HEAP
void     **FastHeapObjects[MAX_HEAPOBJECT_SIZE];
int        FastHeapCount[MAX_HEAPOBJECT_SIZE];
long       FastHeapNewMissCount[MAX_HEAPOBJECT_SIZE];
long       FastHeapDeleteMissCount[MAX_HEAPOBJECT_SIZE];
long       FastHeapNewHitCount[MAX_HEAPOBJECT_SIZE];
long       FastHeapDeleteHitCount[MAX_HEAPOBJECT_SIZE];
#endif

// ***FUNCTIONS*********************************************************************** 

int
CompareParallelIDs(TParallelID *ID1, TParallelID *ID2)

{
    if (*ID1 < *ID2)
        return -1;
    else
        if (*ID1 > *ID2)
            return 1;
        else
            return 0;
}

// =================================================================================== 

int
CompareParallelObjectsByParallelID(TParallelObject *Object1, TParallelObject *Object2)

{
    if (Object1->GetParallelID() > Object2->GetParallelID())
        return 1;
    else 
        if (Object1->GetParallelID() < Object2->GetParallelID())
            return -1;
        else
            return 0;
}

// ***METHODS************************************************************************* 

#ifdef NO_INLINE
#  include "TParallelObject.i.C"
#endif

// =================================================================================== 
#define ParentClass
// =================================================================================== 

TParallelObject::TParallelObject(TClassInfo *theClassInfo, TParallelID theParallelID)

{
    Container = 0;
    ClassInfo = theClassInfo;
    ParallelID = theParallelID;
    Register();
}

// =================================================================================== 

TParallelObject::TParallelObject(TClassInfo *theClassInfo, TBaseStream *Stream,
                                 TTransferInfo *TransferInfo)

{
    Container = 0;
    ClassInfo = theClassInfo;
    ParallelID = INVALID;
    if (!Decode(Stream, TransferInfo))
    {
        Fail("cannot decode");
    }
    if (ClassInfo)
    {    
        TClassInfo *ContainerClass;

        ClassInfo->Insert(this);
        if ((ContainerClass = ClassInfo->GetContainerClass()))
            Container = (TContainerObject *) ContainerClass->GetObjectPtr(ParallelID, FALSE);
    }
}

// =================================================================================== 

TParallelObject::~TParallelObject(void)

{
    Unregister();
}

// =================================================================================== 

int
TParallelObject::Compare(TBaseObject *OtherObject, int CompareCode) const

{
    TParallelObject *OtherParallelObject = (TParallelObject *) OtherObject;

    switch (CompareCode)
    {
    case PO_CC_PARALLEL_ID:

        return CompareLong(GetParallelID(), OtherParallelObject->GetParallelID());

    default:
        
        Warning("Compare: invalid compare mode %d", CompareCode);
    }
    return 0;
}

// =================================================================================== 

bool
TParallelObject::Decode(TBaseStream *Stream, TTransferInfo *TransferInfo)

{
    switch(TransferInfo? TransferInfo->TransferMode : TM_ConstData)
    {
    case TM_ConstData:
        IfDebug(DM_CODE, Print("Decode: const data"));
        return (Stream->Decode(Flags) AND
                Stream->Decode(ParallelID));

    case TM_VarData:
        IfDebug(DM_CODE, Print("Decode: var data"));
        return TRUE;

    case TM_AsciiData:
        IfDebug(DM_CODE, Print("Decode: ascii data"));
        return TRUE;

    default:
        IfDebug(DM_CODE, Print("Decode: TransferMode?"));
    }
    return TRUE;
}

// =================================================================================== 
#ifndef NO_FAST_HEAP
void
TParallelObject::operator delete(void *p, size_t size)

{
    if (size >= MAX_HEAPOBJECT_SIZE)
    {
        free(p);
        return;
    }
    if (FastHeapObjects[size] == 0)
        FastHeapObjects[size] = (void **) malloc(MAX_HEAPOBJECTS * sizeof(void *));
    if (FastHeapCount[size] < MAX_HEAPOBJECTS)
    {
        FastHeapObjects[size][FastHeapCount[size]] = p;
        FastHeapCount[size]++;
        FastHeapDeleteHitCount[size]++;
    }
    else
    {
        FastHeapDeleteMissCount[size]++;
        free(p);
    }
}
#endif
// =================================================================================== 

bool
TParallelObject::DisposeRange(TDistributedRange *)

{
    return TRUE;
}

// =================================================================================== 
#ifndef NO_FAST_HEAP
void
TParallelObject::DumpHeapManagerStat(void)

{
    int i;

    for (i = 0; i <MAX_HEAPOBJECT_SIZE; i++)
        if (FastHeapObjects[i])
            fprintf(stderr, "HEAP size %d: stored:%d * new hits/count:%ld/%ld "
                    "delete hits/count: %ld/%ld\n", i, FastHeapCount[i],
                    FastHeapNewHitCount[i], FastHeapNewHitCount[i] + FastHeapNewMissCount[i],
                    FastHeapDeleteHitCount[i], 
                    FastHeapDeleteHitCount[i] + FastHeapDeleteMissCount[i]);
}
#endif
// =================================================================================== 

bool
TParallelObject::Encode(TBaseStream *Stream, TTransferInfo *TransferInfo) const

{
    switch (TransferInfo? TransferInfo->TransferMode : TM_ConstData)
    {
    case TM_ConstData:
        IfDebug(DM_CODE, Print("Encode: const data"));
        return (Stream->Encode(Flags) AND
                Stream->Encode(ParallelID));

    case TM_VarData:
        IfDebug(DM_CODE, Print("Encode: var data"));
        return TRUE;
        
    case TM_AsciiData:
        IfDebug(DM_CODE, Print("Encode: ascii data"));
        return TRUE;
        
    default:
        IfDebug(DM_CODE, Print("Encode: TransferMode?"));
    }
    return TRUE;
}


// =================================================================================== 

bool
TParallelObject::GetBoundaries(TTimeStep &, TDLListPrimitive *)

{
    return TRUE;
}

// =================================================================================== 

TClassID
TParallelObject::GetClassID(void) const

{
    if (ClassInfo == 0)
    {
        Warning("GetClassID: no class info defined");
        return INVALID;
    }
    return ClassInfo->GetID();
}

// =================================================================================== 

TString
TParallelObject::GetClassName(void) const

{
    return "TParallelObject";
}

// =================================================================================== 

TCPNID
TParallelObject::GetCPNID(void)  const

{
    return INVALID;
}

// =================================================================================== 

TParallelObject *
TParallelObject::GetInfo(TInfoID InfoID, bool Create, bool Load) const

{
    if (Container == 0)
    {
        Error("GetInfo: no container object defined");
        return 0;
    }
    return Container->GetInfo(InfoID, Create, Load);
}

// =================================================================================== 

TString
TParallelObject::GetInstanceName(void) const

{
    static char Format[32] = { 0 };
    char Line[32];

    if (0 == Format[0]) {
        sprintf(Format, "PID:%s %%s", PARALLEL_ID_PRINTF_FORMAT);
    }
    sprintf(Line, Format, ParallelID, IsDistributable()? "dist":"non-dist");
    return Line;
}

// =================================================================================== 

double
TParallelObject::GetLoad(TCPNID, TInfoID, TInfoID)

{
    return 0.0;
}

// =================================================================================== 
#ifndef NO_FAST_HEAP
void
TParallelObject::InitHeapManager(void)

{
    int i;

    for (i = 0; i < MAX_HEAPOBJECT_SIZE; i++)
    {
        FastHeapObjects[i] = 0;
        FastHeapCount[i] = 0;
    }
}
#endif
// =================================================================================== 

bool
TParallelObject::IsInRange(const TDistributedRange *) const

{
    return TRUE;
}

// =================================================================================== 

bool
TParallelObject::IsDistributable(void) const

{
    return TRUE;
}

// =================================================================================== 
#ifndef NO_FAST_HEAP
void *
TParallelObject::operator new(size_t size)

{
    if (size >= MAX_HEAPOBJECT_SIZE)
    {
        fprintf(stderr, "HEAP: class size %d exceding limit %d\n", size, MAX_HEAPOBJECT_SIZE);
        return malloc(size);
    }
    if (FastHeapCount[size])
    {
        FastHeapNewHitCount[size]++;
        FastHeapCount[size]--;
        return FastHeapObjects[size][FastHeapCount[size]];
    }
    else
    {
        FastHeapNewMissCount[size]++;
        return malloc(size);
    }
}
#endif
// =================================================================================== 

bool
TParallelObject::ReadEvent(TStream *Stream)

{
    NotOverloaded("ReadEvent");
    return TRUE;
}

// =================================================================================== 

void
TParallelObject::Register(void)

{
    if (ClassInfo)
    {    
        TClassInfo *ContainerClass;

        ClassInfo->Insert(this);
        if ((ContainerClass = ClassInfo->GetContainerClass()))
            if ((Container = (TContainerObject *) 
                 ContainerClass->GetObjectPtr(ParallelID, FALSE)) == 0) /* -> */  ;
                //               ContainerClass->GetObjectPtr(ParallelID, TRUE)) == 0)
                //              Warning("CONSTRUCTOR: cannot retrieve pointer to container");
    }    
}

// =================================================================================== 

void
TParallelObject::SetParallelID(TParallelID theParallelID)

{
    Unregister();
    ParallelID = theParallelID;
    Register();
}

// =================================================================================== 

void
TParallelObject::Unregister(void)

{
    if (ClassInfo)
        ClassInfo->Delete(this);
}

// =================================================================================== 

bool
TParallelObject::Write(TBaseStream *Stream)

{
    return Stream->Encode(ParallelID);
}

// =================================================================================== 
#undef ParentClass
