/*      file     : pvm/TPVMParallelControl.C
        project  : Parallel Toolbox 2.x
        author   : Marcus Rickert
        RCS      : $Id: TPVMParallelControl.C,v 1.23 2005/03/29 20:32:32 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/types.h>

#include <Error.h>
#include <File.h>
#include <TEnvironment.h>
#include <TConfiguration.h>

#include <ParallelTypes.h>
#include <TCPNInfo.h>
#include <THostInfo.h>

#include "Pvm.h"
#include "PVMConfig.h"
#include "LocalPvm.h"
#include "TPVMParallelControl.h"

// ***DEFINES*************************************************************************

#define PVM_HOST_FILENAME           "hostfile"
#define PVM_HOST_MASK                0x8FFF0000

// ***CONSTANTS***********************************************************************

char *
PvmRoutingModes[] =

{
    "unknown",
    "DontRoute",
    "AllowDirect",
    "RouteDirect"
};

// ***METHODS************************************************************************* 

// =================================================================================== 
#define ParentClass TParallelControl
// =================================================================================== 

TPVMParallelControl::TPVMParallelControl(void)

{
    _Init();
    debug(Print(DM_LIFESPAN, "CONSTRUCTOR"));
}

// =================================================================================== 

TPVMParallelControl::~TPVMParallelControl(void)

{
    Disenroll();
    DeleteObject(PvmMutable);
    if (InactiveHosts)
    {
        InactiveHosts->DeleteAll(TRUE);
        DeleteObject(InactiveHosts);
    }
    debug(Print(DM_LIFESPAN, "Deactivate"));
}

// =================================================================================== 

bool
TPVMParallelControl::Activate(void)

{
    TConfiguration *C = Configuration;

    debug(Print(DM_LIFESPAN, "Activate"));
    RoutingMode = (int) C->GetLong(PVM_GROUP, PVM_NAME_ROUTING_MODE);
    if (C->Exists(PVM_GROUP, PVM_NAME_HOSTFILENAME))
        HostFilename = C->GetString(PVM_GROUP, PVM_NAME_HOSTFILENAME);
    if (!Enroll())
        return FALSE;
    if (IsMaster())
        if (!ReadHostfile())
            return FALSE;
    return TRUE;
}

// =================================================================================== 

bool
TPVMParallelControl::AddCPN(THostInfo *aHostInfo, int &NrOfInstances)

{
    int       SpawnFlags;
    int       ToBeSpawned;
    char     *Executable;
    int       i;
    int      *ChildTIDs = 0;
    TCPNInfo *CPNInfo = 0;

    ToBeSpawned = NrOfInstances;
    NrOfInstances = 0;
    if (ToBeSpawned <= 0)
        return TRUE;    
    if (!CheckCPNInfoList())
        return FALSE;
    Executable = (char *) Environment->GetExecutable();
    SpawnFlags = PvmTaskHost;
    Print(DM_CONTROL, "adding %d CPN on host %s running '%s'", ToBeSpawned,
          (char *) aHostInfo->Name(), Executable);
    ChildTIDs = (int *) calloc(ToBeSpawned, sizeof(int));
    if ((NrOfInstances = pvm_spawn(Executable, 0, SpawnFlags, (char *) aHostInfo->Name(),
                                   ToBeSpawned, ChildTIDs)) < 0)
    {
        Error("AddCPN: pvm_spawn returns %i", NrOfInstances);
        return FALSE;
    }
    if (NrOfInstances < ToBeSpawned)
        Warning("fewer instances (%d) spawned than requested", NrOfInstances);
    for (i = 0; i < ToBeSpawned; i++)
    {
        if (ChildTIDs[i] < 0)
        {
            Error("AddCPN: child #%d reports error %d", i, ChildTIDs[i]);
            return FALSE;
        }
        CPNInfo = GetCPNInfo(aHostInfo, ChildTIDs[i]);
        Mutable->CPNInfoList->Append(CPNInfo);
        Print(DM_CONTROL, "added CPN %X (%s)", CPNInfo->ID, (char *) CPNInfo->Name);
    }    
    Free(ChildTIDs);
    return TRUE;
}

/* =================================================================================== */

bool
TPVMParallelControl::CheckCPNInfoList(void) const

{
    if (Mutable->CPNInfoList == 0)
    {
        if (!CheckHostInfoList())
            return FALSE;
        Mutable->LocalCPNInfo = GetCPNInfo(Mutable->LocalHostInfo, LocalCPNID);
        Environment->SetProcessName(Mutable->LocalCPNInfo->Name);
        Mutable->CPNInfoList = new TCPNInfoList;
        Mutable->CPNInfoList->Append(Mutable->LocalCPNInfo);
    }
    return TRUE;
}

/* =================================================================================== */

bool
TPVMParallelControl::CheckHostInfoList(void) const

{
    int i;

    if (Mutable->HostInfoList == 0)
    {
        if (!GetPvmHostInfos())
            return FALSE;
        Mutable->HostInfoList = new THostInfoList;
#ifdef PVM_FAKE_MASTER
        if (!FakeMasterHostInfo(PvmMutable->PvmHostInfos))
            return FALSE;
#endif
        for (i = 0; i < PvmMutable->PvmNrOfHosts; i++)
        {
            THostInfo *HostInfo = 0;

            if ((HostInfo = GetHostInfo(PvmMutable->PvmHostInfos[i].hi_name)) == 0)
                return FALSE;
            HostInfo->ID = PvmTID(PvmMutable->PvmHostInfos + i);
            HostInfo->Performance = PvmMutable->PvmHostInfos[i].hi_speed;
            HostInfo->ArchName = PvmMutable->PvmHostInfos[i].hi_arch;
            HostInfo->ArchCode = PvmArchCode(PvmMutable->PvmHostInfos[i].hi_arch);
#ifdef DEBUG
            HostInfo->DebugMask = Debug.GetMask();
#endif
            Mutable->HostInfoList->Append(HostInfo);
            if (CPNonHost(LocalCPNID, HostInfo->ID))
            {
                Mutable->LocalHostInfo = HostInfo;
                Environment->SetHostName(HostInfo->GetShortName());
            }
        }
        if (Mutable->LocalHostInfo == 0)
        {
            Error("CheckHostInfoLists: no local hostinfo found");
            return FALSE;
        }
    }
    return TRUE;
}

/* =================================================================================== */

bool
TPVMParallelControl::CPNonHost(TCPNID aCPNID, THostID aHostID) const

{
    return (THostID) (aCPNID & PVM_HOST_MASK) == aHostID;
}

/* =================================================================================== */

bool
TPVMParallelControl::Disenroll(void)

{
    int PvmError;

    if (IsEnrolled())
    {
        Print(DM_CONTROL, "Disenroll from PVM");
        if ((PvmError = pvm_exit()) < 0)
        {
            Error("Disenroll: pvm_exit returns: %i\n", PvmError);
            return FALSE;
        }
        if (!ParentClass::Disenroll())
            return FALSE;
    }
    return TRUE;
}

// =================================================================================== 

bool
TPVMParallelControl::Enroll(void)

{
    const char *ArchName;

    if (!IsEnrolled())
    {
        if ((LocalCPNID = PvmMyTID()) < 0)
        {
            Error("Enroll: PVM not installed/activated!");
            LocalErrorStatus = Status_CommunicationError;
            return FALSE;
        }
        pvm_setopt(PvmRoute, RoutingMode);
        PvmCatchOut(stdout);
        ParentCPNID = pvm_parent();
        Print("Enroll into PVM with TID %X", LocalCPNID);
        Environment->SetLocalID(LocalCPNID);
        if ((ArchName = getenv("PVM_ARCH")))
            Environment->SetArchitecture(ArchName);
        if (IsMaster() OR SlavesNeedPrintPrefix())
        {
            char Prefix[32];

            sprintf(Prefix, "[T%lx] ", LocalCPNID);
            Environment->SetPrintPrefix(Prefix);
        }
        else
            Environment->SetPrintPrefix("");
        if (!ParentClass::Enroll())
            return FALSE;
    }
    return TRUE;
}

/* =================================================================================== */

bool
TPVMParallelControl::FakeMasterHostInfo(struct pvmhostinfo *aHostInfo) const

{
    THostInfo *HostInfo = 0;

    if ((HostInfo = GetHostInfo("Master")) == 0)
        return FALSE;
    if (IsMaster())
        HostInfo->ID = LocalCPNID;
    else
        HostInfo->ID = pvm_parent();
    Print("hi_tid=%X", HostInfo->ID);
    
    HostInfo->Performance = aHostInfo->hi_speed;
    HostInfo->ArchName = aHostInfo->hi_arch;
    HostInfo->ArchCode = PvmArchCode(aHostInfo->hi_arch);
#ifdef DEBUG
    HostInfo->DebugMask = Debug.GetMask();
#endif
    Mutable->HostInfoList->Append(HostInfo);
    if (IsMaster())
    {
        Mutable->LocalHostInfo = HostInfo;
        Environment->SetHostName(HostInfo->GetShortName());
    }
    return TRUE;
}

// =================================================================================== 

TString
TPVMParallelControl::GetClassName(void) const

{
    return "TPVMParallelControl";
}

// =================================================================================== 

THostInfo *
TPVMParallelControl::GetHostInfo(const char *Hostname) const

{
    THostInfo     *HostInfo = 0;
    THostInfoEnum  CHost;

    CHost.WalksForwardThrough(InactiveHosts);
    while (CHost())
    {
        if (CHost->SameName(Hostname))
        {
            HostInfo = CHost;
            // we have to cling to the spelling defined by PVMD:
            HostInfo->Name = Hostname;
            InactiveHosts->Delete(HostInfo);
            return HostInfo;
        }
        CHost++;
    }
    if ((HostInfo = new THostInfo(Hostname)) == 0)
    {
        Error("cannot allocate host info");
        return 0;
    }
    return HostInfo;
}

// =================================================================================== 

TCPNID
TPVMParallelControl::GetParentCPNID(void) const

{
    if (ParentCPNID == PvmNoParent)
        return LocalCPNID;
    else
        return ParentCPNID;
}

// ===================================================================================

bool
TPVMParallelControl::GetPvmHostInfos(void) const

{
    int PvmError;

    if (PvmMutable->PvmHostInfos == 0)
        if ((PvmError = pvm_config(&PvmMutable->PvmNrOfHosts, 
                                   &PvmMutable->PvmNrOfArchs, 
                                   &PvmMutable->PvmHostInfos)) < 0)
        {
            Error("GetPvmHostInfos: pvm_config returns %i!\n", PvmError);
            return FALSE;
        }
    return TRUE;
}

// ===================================================================================

void
TPVMParallelControl::_Init(void)

{
    TConfiguration *C = Configuration;

    InactiveHosts = new TDLList<THostInfo>;
    PvmMutable = new TPvmMutable;
    PvmMutable->PvmHostInfos = 0;    
    ParentCPNID = INVALID;
    LocalCPNID = INVALID;
    C->Set(TConfigEntry(PVM_GROUP, PVM_NAME_ROUTING_MODE, 
                        (long) PVM_DEFAULT_ROUTING_MODE));
}

// ===================================================================================

bool
TPVMParallelControl::IsMaster(void) const

{
    return (ParentCPNID == PvmNoParent);
}

// ===================================================================================

bool
TPVMParallelControl::MarkTopologyChanged(void)

{
    PvmMutable->PvmHostInfos = 0;
    return TRUE;
}

/* =================================================================================== */

bool
TPVMParallelControl::ReadHostfile(void)

{
    char         Filename[256];
    FILE        *input;
    char         Line[256];
    int          First, Last;
    THostInfo   *NewInfo = 0;
    
    Filename[0] = 0;
    if (HostFilename.IsDefined())
        strcpy(Filename, HostFilename);
    else
    {
        if (fexist(PVM_HOST_FILENAME))
            strcpy(Filename, PVM_HOST_FILENAME);
        else
        {
            strcpy(Filename, Environment->GetHomeFilename(PVM_HOST_FILENAME));
            if (fexist(Filename))
                Filename[0] = 0;
        }
    }
    if (Filename[0])
    {
        TString ApplicationFilename = Environment->GetApplicationFilename(Filename);

        if ((input = fopen(ApplicationFilename, "r")) == 0)
        {
            Error("cannot open file '%s'", (char *) ApplicationFilename);
            return FALSE;
        }
        while (!feof(input))
        {
            Line[0] = 0;
            fgets(Line, 256, input);
            if (Line[0])
            {
                if (Line[0] == '#' AND Line[1] == '!')
                {
                    First = 2;
                    while (Line[First] AND (Line[First] == ' ' OR Line[First] == '\t'))
                        First++;
                    Last = First;
                    while (Line[Last] AND Line[Last] != ' ' AND Line[Last] != '\t')
                        Last++;
                    Line[Last] = 0;                 
                    NewInfo = new THostInfo(Line + First);
                    NewInfo->MaxNrOfCPN = atoi(Line + Last + 1);
                    InactiveHosts->Append(NewInfo);
                    // Print("HostInfo: %s", (char *) NewInfo->GetInstanceName());
                }
            }
        }
        Print(DM_CONTROL, "%ld info(s) read from hostfile %s", 
              InactiveHosts->GetNrOfElements(), (char *) ApplicationFilename);
        fclose(input);
    }
    else
        Warning("no hostfile found");
    return TRUE;
}

// ===================================================================================
#undef ParentClass
